# _*_ coding:UTF-8 _*_

# Author: Bing Shi
# Institution: City University of Hong Kong
# Date: 2017.4~2018.3
# Function： Generate g1 and g2, then make features of them. 
# Using:

import  os
import  sys
import  copy
import  math
import  matplotlib.pyplot as plt
import  numpy as np  
from  numpy import random as nr

# === setting of parameter ===
k = 10                 

samples_num = 100
outliers_num = 5
sigma = 0.1  
mu = 0
multiple = 100
a = multiple*1.1
b = multiple*1.3

def gengraph(f1, f2, samples_num = samples_num, outliers_num = outliers_num, sigma = sigma, mu = mu):    
    np.set_printoptions(precision=3)
    firstArray = nr.random(size=(samples_num, 2))*multiple          # G1   
    secondArray = copy.deepcopy(firstArray)                         # G2
    if sigma != 0:
        deformationArray = nr.normal(mu, sigma, size=(samples_num, 2))*multiple/10  
        secondArray = secondArray + deformationArray 
    if outliers_num != 0:    
        outlierArray = (b - a) * nr.random_sample(size=(outliers_num, 2)) + a 
        secondArray = np.vstack((secondArray, outlierArray))    

    np.savetxt(f1, firstArray, fmt='%.4f %.4f', newline='\r\n')     # save data
    np.savetxt(f2, secondArray, fmt='%.4f %.4f', newline='\r\n')
    return (firstArray, secondArray)
#end

def drawgraph(firstArray, secondArray):    
    ''' plot scatter    #matplotlib.rc('font', size=6) '''
    plt.subplot(121)
    plt.title('G1')
    plt.plot(firstArray[:,0], firstArray[:,1],'+')           
    
    plt.subplot(122)
    plt.title('G2')
    plt.plot(secondArray[:,0], secondArray[:,1],'+')
    
    plt.savefig('g1g2.png', dpi=125)    
    plt.show()  
#end

def makefeatures(fn):
    mname, ext = os.path.splitext(fn)
    points = np.loadtxt(fn, dtype=np.float32)        #load keypoints
    n = points.shape[0]
    print('Row: %d'%n)
    x1, x2 = np.meshgrid(points[:, 0], points[:, 0])
    y1, y2 = np.meshgrid(points[:, 1], points[:, 1])
    deltax = x2-x1
    deltay = y2-y1
    distances=np.sqrt(deltax**2 + deltay**2)
    neighbours = distances.argsort()  
    features = []
    for i in range(n):
        index = neighbours[i][1:k]
        print('%d:\nindex:'%i, end=' ');   print(index)
        d = distances[i][index]
        print('distances: ', end=' ');     print(d)
        min=np.min(d);   assert min==distances[i][neighbours[i][1]]
        max=np.max(d);   assert max==distances[i][neighbours[i][k-1]]
        mean=np.mean(d)  
        std=np.std(d)
        median=np.median(d)
        p1=np.percentile(d,25)
        p3=np.percentile(d,75)
        f,bins = np.histogram(d,bins=[min, p1, median, p3, max])          #f:frequency
        feat=[10*min/max, 10*mean/max, 10*std/max]                        #feat = feat + list(f)
        print('near,far: %d %d'%(neighbours[i][1], neighbours[i][k-1]))
        x=[deltax[i][neighbours[i][1]], deltay[i][neighbours[i][1]]]
        y=[deltax[i][neighbours[i][k-1]], deltay[i][neighbours[i][k-1]]]
        lx=np.sqrt(np.dot(x, x))
        ly=np.sqrt(np.dot(y, y))
        cos_angle=np.dot(x, y)/(lx*ly)
        cos_angle = np.maximum(-1,cos_angle)
        cos_angle = np.minimum(1,cos_angle)
        angle=np.arccos(cos_angle)
        feat.append(angle)
        features.append(feat)
        print(feat)
    features = np.array(features)   
    np.savetxt(mname+'.des.txt', features, fmt="%f", delimiter=' ', newline='\r\n') 
    plt.hist(features)       
    plt.show()
#end 

if __name__=='__main__':  
    (fa, sa) = gengraph('g1.txt', 'g2.txt')
    drawgraph(fa,sa)
    makefeatures('g1.txt')
    makefeatures('g2.txt')



# 

# Author: Bing SHI
# Institution: City University of Hong Kong
# Date: 2017.7~2017.12
# Function： Data preproccessing: Remove reduplicative row in two graphs for graph matching.
# Using: python preproc.py fm1 fm2
# fm1, fm2 are respectively main names of two graphs.

import os                      
import sys
import numpy as np

def cleanout(keyp, keyflag):
    num = keyp.shape[0] 
    for i in range(num):
        if keyflag[i] == 0: continue
        li = list(keyp[i])
        for j in range(i+1, num): 
            if li == list(keyp[j]): keyflag[j] = 0
# end of cleanout

def remrr(f1, f2):
    '''
    Function： Data preproccessing: Remove reduplicative row in two graphs for graph matching.
    Using: python preproc.py fm1 fm2
    here, fm1, fm2 are respectively main names of two graphs. Every graph consist of fmx.kps.txt and fmx.des.txt.
    '''
    keyp1 = np.loadtxt(f1+'.kps.txt', dtype=np.float32)
    feat1 = np.loadtxt(f1+'.des.txt', dtype=np.int16)
    keyp2 = np.loadtxt(f2+'.kps.txt', dtype=np.float32)
    feat2 = np.loadtxt(f2+'.des.txt', dtype=np.int16)
    if keyp1.shape[0] != keyp2.shape[0]: print('data error!'); return
    
    keyflag1 = np.ones(keyp1.shape[0], np.int16)
    keyflag2 = np.ones(keyp2.shape[0], np.int16)
    cleanout(keyp1, keyflag1)
    cleanout(keyp2, keyflag2)
    
    keyflag = np.ones(keyp1.shape[0], np.int16)   
    for i in range(len(keyflag1)):
        if keyflag1[i] == 0 or keyflag2[i] == 0: keyflag[i] = 0
       
    keyp12=keyp1[keyflag == 1];     feat12=feat1[keyflag == 1]
    keyp22=keyp2[keyflag == 1];     feat22=feat2[keyflag == 1]
    print(f1+':%d %d'%(len(keyp12), keyflag1.sum()));   print(f2+':%d %d'%(len(keyp22), keyflag2.sum()));   print('con: %d'% keyflag.sum())

    np.savetxt(f1+'.kps.1.txt', keyp12, fmt="%f", delimiter=' ', newline='\r\n')
    np.savetxt(f1+'.des.1.txt', feat12, fmt="%u", delimiter=' ', newline='\r\n')
    np.savetxt(f2+'.kps.1.txt', keyp22, fmt="%f", delimiter=' ', newline='\r\n')
    np.savetxt(f2+'.des.1.txt', feat22, fmt="%u", delimiter=' ', newline='\r\n')
# End     

if __name__ == "__main__":
    remrr(sys.argv[1], sys.argv[2])  



#!/usr/bin/python
# _*_ encoding: UTF-8 _*_

from distutils.core import setup        

setup(  name = 'hgm',
        version = '0.5.3',
        packages=['hgm'], 
        py_modules = ['kdfe','preproc','ggmf'],
        description = 'Hyper-graph matching API', 
        author = 'Bing SHI',            
        author_email = 'bing.shi@aliyun.com',
        url = 'https://github.com/bing-shi/HGM',
)
  



# 

# Author: Bing SHI
# Institution: City University of Hong Kong
# Date: 2017.7~2017.12
# Function： KeyPoint detection and features extraction for graph matching.
# Using: python kdfe.py fm1 fm2    
# fm1, fm2 are respectively main name of two image files.

import os
import sys
import numpy as np
import cv2
from matplotlib import pyplot as plt
from copy import deepcopy

# parameters setting
RATIO = 0.80        # According to the need to change.

def siftdetext(f1, f2):
    '''
    Function： KeyPoint detection and features extraction for graph matching.
    Using: python kdfe.py fm1 fm2    
    here, fm1, fm2 are respectively main name of two image files.
    '''
    MIN_MATCH_COUNT = 10  
    fn1, ext = os.path.splitext(f1)
    fn2, ext = os.path.splitext(f2)
    #--1-- read images
    img1 = cv2.imread(f1,0)          # queryImage
    img2 = cv2.imread(f2,0)          # trainImage   
    #--2-- detecte keypoints and extract SIFT features    
    sift = cv2.xfeatures2d.SIFT_create()          
    # find the keypoints and descriptors with SIFT
    kp1, des1 = sift.detectAndCompute(img1,None)  
    kp2, des2 = sift.detectAndCompute(img2,None)
    #--3-- matching
    FLANN_INDEX_KDTREE = 0
    index_params = dict(algorithm = FLANN_INDEX_KDTREE, trees = 5)
    search_params = dict(checks = 50)
    flann = cv2.FlannBasedMatcher(index_params, search_params)
    matches = flann.knnMatch(des1,des2,k=2)
    #--4-- ratio test.   
    good = []
    for m,n in matches:
        if m.distance < RATIO * n.distance:
            good.append(m) 
    print('good matching: %d'%len(good))    
    # Homography   
    if len(good) > MIN_MATCH_COUNT:
        src_pts = np.float32([ kp1[m.queryIdx].pt for m in good ]).reshape(-1,1,2)
        dst_pts = np.float32([ kp2[m.trainIdx].pt for m in good ]).reshape(-1,1,2)
        M, mask = cv2.findHomography(src_pts, dst_pts, cv2.RANSAC,5.0)
        matchesMask = mask.ravel().tolist()                # corresponding to good
        h,w = img1.shape                                   # h,w,d = img1.shape for color image
        pts = np.float32([ [0,0],[0,h-1],[w-1,h-1],[w-1,0] ]).reshape(-1,1,2)
        dst = cv2.perspectiveTransform(pts,M)
        #img2 = cv2.polylines(img2,[np.int32(dst)],True,255,3, cv2.LINE_AA)
    else:
        print("Not enough matches are found - %d/%d" % (len(good),MIN_MATCH_COUNT))
        matchesMask = None       
    #--5-- draw matches
    draw_params = dict(matchColor = (0,255,0),     # draw matches in green color  
                       singlePointColor = None,
                       matchesMask = matchesMask,  # draw only inliers
                       flags = 2)
    img3 = cv2.drawMatches(img1,kp1,img2,kp2,good,None,**draw_params)
    plt.imshow(img3, 'gray');     plt.show()
    #--6-- save results--
    mm=np.asarray(matchesMask)
    src = [[kp1[m.queryIdx].pt[0],kp1[m.queryIdx].pt[1]] for m in good ] 
    srcarr = np.asarray(src)
    srcarr1 = srcarr[mm==1]
    np.savetxt(fn1+'.kps.txt',srcarr1,fmt='%f %f',newline='\r\n')
    desind1 = [m.queryIdx for m in good ]
    des12 = des1[desind1]
    desi1 = des12.astype(int)
    desi3 = desi1[mm==1]
    np.savetxt(fn1+'.des.txt',desi3,fmt='%d '*128,newline='\r\n') 
    
    dst = [[kp2[m.trainIdx].pt[0],kp2[m.trainIdx].pt[1]] for m in good ]
    dstarr= np.asarray(dst)
    dstarr1 = dstarr[mm==1]
    np.savetxt(fn2+'.kps.txt',dstarr1,fmt='%f %f',newline='\r\n')    
    desind2 = [m.trainIdx for m in good ]
    des22 = des2[desind2]    
    desi2=des22.astype(int)   
    desi4 = desi2[mm==1]
    np.savetxt(fn2+'.des.txt',desi4,fmt='%d '*128,newline='\r\n') 
# End

if __name__ == "__main__":
    siftdetext(sys.argv[1], sys.argv[2])
